#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# utilities.py
"""Common functions for the DVHA-Stats."""
#
# Copyright (c) 2020 Dan Cutright
# This file is part of DVHA-Stats, released under a MIT license.
#    See the file LICENSE included with this distribution, also
#    available at https://github.com/cutright/DVHA-Stats

import numpy as np
from os.path import isfile, splitext


def apply_dtype(value, dtype):
    """Convert value with the provided data type

    Parameters
    ----------
    value : any
        Value to be converted
    dtype : callable
        python reserved types, e.g., int, float, str, etc. However, dtype
        could be any callable that raises a ValueError on failure.

    Returns
    ----------
    any
        The return of dtype(value) or numpy.nan on ValueError
    """
    if dtype is None:
        return value
    try:
        value = dtype(value)
    except ValueError:
        value = np.nan
    return value


def csv_to_dict(csv_file_path, delimiter=",", dtype=None, header_row=True):
    """Read in a csv file, return data as a dictionary

    Parameters
    ----------
    csv_file_path : str
        File path to the CSV file to be processed.
    delimiter : str
        Specify the delimiter used in the csv file (default = ',')
    dtype : None, Type, optional
        Optionally force values to a type (e.g., float, int, str, etc.).
    header_row : bool, optional
        If True, the first row is interpreted as column keys, otherwise row
        indices will be used

    Returns
    -------
    dict
        CSV data as a dict, using the first row values as keys
    """

    with open(csv_file_path) as fp:

        # Read first row, determine column keys
        first_row = fp.readline().strip().split(delimiter)
        if header_row:
            keys = first_row
            data = {key: [] for key in keys}
        else:
            keys = list(range(len(first_row)))
            data = {key: [apply_dtype(first_row[key], dtype)] for key in keys}

        # Iterate through remaining rows, append values to data
        for r, line in enumerate(fp):
            row = line.strip().split(",")
            for c, value in enumerate(row):
                data[keys[c]].append(apply_dtype(value, dtype))

    return data


def dict_to_array(data, key_order=None):
    """Convert a dict of data to a numpy array

    Parameters
    ----------
    data : dict
        File path to the CSV file to be processed.
    key_order : None, list of str
        Optionally the order of columns

    Returns
    -------
    dict
        A dictionary with keys of 'data' and 'columns', pointing to a
        numpy array and list of str, respectively
    """
    var_names = key_order if key_order is not None else list(data.keys())
    arr_data = [data[key] for key in var_names]
    return {"data": np.asarray(arr_data).T, "var_names": var_names}


def import_data(data, var_names=None):
    """Generalized data importer for np.ndarray, dict, and csv file

    Parameters
    ----------
    data : numpy.array, dict, str
        Input data (2-D) with N rows of observations and
        p columns of variables.  The CSV file must have a header row
        for column names.
    var_names : list of str, optional
        If data is a numpy array, optionally provide the column names.

    Returns
    ----------
    np.ndarray, list
        A tuple: data as an array and variable names as a list
    """
    if isinstance(data, np.ndarray):
        var_names = (
            var_names if var_names is not None else list(range(data.shape[1]))
        )
        return data, var_names
    if isinstance(data, dict):
        data = dict_to_array(data)
        return data["data"], data["var_names"]
    if isfile(data):
        if splitext(data)[1] == ".csv":
            data = dict_to_array(csv_to_dict(data, dtype=float))
            return data["data"], data["var_names"]

    msg = "Invalid data provided - must be a numpy array, dict, or .csv file"
    raise NotImplementedError(msg)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# _version.py
"""Package initialization for DVHA-Stats."""
# Copyright (c) 2020 Dan Cutright
# This file is part of dicompyler-core, relased under a MIT license.
#    See the file LICENSE included with this distribution, also
#    available at https://github.com/cutright/DVHA-Stats

__author__ = "Dan Cutright"
__email__ = "dan.cutright@gmail.com"
__version__ = "0.1.7"
__release__ = "0.1.7"
__version_info__ = (0, 1, 7)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# test_stats.py
"""unittest cases for stats."""
#
# Copyright (c) 2020 Dan Cutright
# This file is part of DVHA-Stats, released under a MIT license.
#    See the file LICENSE included with this distribution, also
#    available at https://github.com/cutright/DVHA-Stats


import unittest
from os.path import join
import numpy as np
from numpy.testing import assert_array_equal, assert_array_almost_equal
from dvhastats import stats
from dvhastats.utilities import import_data
import warnings
from copy import deepcopy


basedata_dir = join("tests", "testdata")
example_data = join(basedata_dir, "multivariate_data_small.csv")
example_data_full = join(basedata_dir, "multivariate_data.csv")


class TestStats(unittest.TestCase):
    """Unit tests for Stats."""

    def setUp(self):
        """Setup files and base data for utility testing."""
        self.data_path = example_data
        self.data_path_full = example_data_full

        data = [
            [np.nan, 48.1, 48.3, 65.1, 47.1, 49.9, 49.5, 48.9, 35.5, 44.5],
            [51.9, 44.3, 44.5, 58.7, 41.1, 43.9, 43.9, 44.5, 31.1, 40.5],
            [48.5, 38.5, 37.1, 53.9, 27.1, 40.7, 34.5, 41.7, 28.7, 37.7],
            [33.9, 21.5, 20.1, 48.7, 12.1, 38.5, 13.9, 33.7, 25.3, 28.1],
            [19.1, 10.9, 9.9, 42.3, 4.3, 36.5, 6.1, 16.9, 13.9, 13.3],
            [12.7, 4.7, 3.9, 31.3, 3.1, 29.1, 3.7, 4.9, 6.9, 4.7],
        ]
        keys = ["V%s" % (i + 1) for i in range(6)]

        self.expected_dict = {key: data[i] for i, key in enumerate(keys)}
        self.expected_dict_nh = {i: row for i, row in enumerate(data)}
        self.expected_arr = np.array(data).T
        self.expected_arr_no_nan = deepcopy(self.expected_arr)
        self.expected_arr_no_nan[0, 0] = 56.5
        self.expected_var_names = keys

        self.const_data = deepcopy(self.expected_arr)
        self.const_data[:, 0] = np.ones_like(self.expected_arr[:, 0])

        self.data, self.var_names = import_data(self.expected_dict)

    def test_arr_import(self):
        """Test array importing"""
        data, var_names = import_data(
            self.expected_arr, self.expected_var_names
        )
        assert_array_equal(data, self.expected_arr)
        self.assertEqual(var_names, self.expected_var_names)

    def test_dict_import(self):
        """Test dict importing"""
        data, var_names = import_data(self.expected_dict)
        assert_array_equal(data, self.expected_arr)
        self.assertEqual(var_names, self.expected_var_names)

    def test_csv_import(self):
        """Test csv importing"""
        data, var_names = import_data(self.data_path)
        assert_array_equal(data, self.expected_arr)
        self.assertEqual(var_names, self.expected_var_names)

    def test_invalid_data_import(self):
        with self.assertRaises(NotImplementedError):
            import_data("test")

    def test_correlation_matrix_type_failure(self):
        """Check that error is raised if corr_type Spearman or Pearson"""
        with self.assertRaises(NotImplementedError):
            stats.CorrelationMatrix(self.data, corr_type="test")

    def test_pearson_r_matrix(self):
        """Test Pearson-R matrix calculation"""
        exp_r = np.array(
            [
                [
                    1.0,
                    0.99242287,
                    0.84208524,
                    0.56410877,
                    0.63275247,
                    0.6431104,
                ],
                [
                    0.99242287,
                    1.0,
                    0.88851061,
                    0.58486366,
                    0.58064696,
                    0.5791157,
                ],
                [
                    0.84208524,
                    0.88851061,
                    1.0,
                    0.83404009,
                    0.74483928,
                    0.66846345,
                ],
                [
                    0.56410877,
                    0.58486366,
                    0.83404009,
                    1.0,
                    0.93433324,
                    0.83444095,
                ],
                [
                    0.63275247,
                    0.58064696,
                    0.74483928,
                    0.93433324,
                    1.0,
                    0.97058164,
                ],
                [
                    0.6431104,
                    0.5791157,
                    0.66846345,
                    0.83444095,
                    0.97058164,
                    1.0,
                ],
            ]
        )
        exp_p = np.array(
            [
                [
                    0.00000000e00,
                    1.23765714e-07,
                    4.39932805e-03,
                    1.13622389e-01,
                    6.74147444e-02,
                    6.16994726e-02,
                ],
                [
                    1.23765714e-07,
                    0.00000000e00,
                    5.89647474e-04,
                    7.57410908e-02,
                    7.83977572e-02,
                    7.93772216e-02,
                ],
                [
                    4.39932805e-03,
                    5.89647474e-04,
                    0.00000000e00,
                    2.70253622e-03,
                    1.34485527e-02,
                    3.45956064e-02,
                ],
                [
                    1.13622389e-01,
                    7.57410908e-02,
                    2.70253622e-03,
                    0.00000000e00,
                    7.51138400e-05,
                    2.67788998e-03,
                ],
                [
                    6.74147444e-02,
                    7.83977572e-02,
                    1.34485527e-02,
                    7.51138400e-05,
                    0.00000000e00,
                    3.16254737e-06,
                ],
                [
                    6.16994726e-02,
                    7.93772216e-02,
                    3.45956064e-02,
                    2.67788998e-03,
                    3.16254737e-06,
                    0.00000000e00,
                ],
            ]
        )

        corr_mat = stats.CorrelationMatrix(self.data)
        warnings.filterwarnings("ignore")
        data = corr_mat.chart_data
        warnings.filterwarnings("default")
        assert_array_almost_equal(data["corr"], exp_r)
        assert_array_almost_equal(data["p"], exp_p)

        exp_norm = np.array(
            [
                5.560821491105694,
                2.168424114591037,
                0.4849069523738952,
                0.36635238844506546,
                3.997163488853074,
                5.497952110420669,
            ]
        )
        exp_norm_p = np.array(
            [
                0.06201303056894888,
                0.3381681380630833,
                0.7847002539774257,
                0.8326214397249772,
                0.1355273594307197,
                0.06399335333154248,
            ]
        )
        assert_array_almost_equal(data["norm"], exp_norm)
        assert_array_almost_equal(data["norm_p"], exp_norm_p)

        self.validate_dict_is_json_compat(data)

    def test_spearman_correlation_matrix(self):
        """Test Spearman correlation matrix calculation"""
        corr_mat = stats.CorrelationMatrix(self.data, corr_type="Spearman")
        self.assertEqual(corr_mat.corr_type, "spearman")

        warnings.filterwarnings("ignore")
        data = corr_mat.chart_data
        warnings.filterwarnings("default")
        self.validate_dict_is_json_compat(data)

    def test_control_chart(self):
        """Test univariate control chart creation and values"""
        ucc = stats.ControlChart(self.data[:, 0])
        str_rep = (
            "center_line: 48.544\ncontrol_limits: "
            "28.199, 68.890\nout_of_control: []"
        )
        self.assertEqual(str(ucc), str_rep)
        self.assertEqual(repr(ucc), str_rep)

        chart_data = ucc.chart_data
        self.assertEqual(len(chart_data["out_of_control"]), 0)
        self.assertEqual(round(chart_data["center_line"], 3), 48.544)
        self.assertEqual(round(chart_data["lcl"], 3), 28.199)
        self.assertEqual(round(chart_data["ucl"], 3), 68.89)
        self.assertEqual(len(ucc.out_of_control_high), 0)
        self.assertEqual(len(ucc.out_of_control_low), 0)

        self.validate_dict_is_json_compat(ucc.chart_data)

    def test_control_chart_with_limits(self):
        """Test univariate control chart creation and values"""
        ucc = stats.ControlChart(
            self.data[:, 0],
            lcl_limit=30,
            ucl_limit=50,
        )
        lcl, ucl = ucc.control_limits
        self.assertEqual(round(lcl, 3), 30)
        self.assertEqual(round(ucl, 3), 50)
        self.assertEqual(len(ucc.out_of_control), 1)

    def test_control_chart_with_all_nan_input(self):
        data = np.array([[np.nan] * 5]).T
        ucc = stats.ControlChart(data)
        self.assertTrue(np.isnan(ucc.center_line))

    def test_hotelling_t2(self):
        """Test multivariate control chart creation and values"""
        ht2 = stats.HotellingT2(self.data, const_policy="omit")

        chart_data = ht2.chart_data
        lcl, ucl = ht2.control_limits
        self.assertEqual(round(chart_data["center_line"], 3), 5.614)
        self.assertEqual(round(lcl, 3), 0)
        self.assertEqual(round(ucl, 3), 7.834)
        self.assertEqual(len(chart_data["out_of_control"]), 0)
        str_rep = (
            "Q: [nan nan nan nan nan nan nan nan nan nan]\n"
            "center_line: 5.614\n"
            "control_limits: 0, 7.834\n"
            "out_of_control: []"
        )
        self.assertEqual(str(ht2), str_rep)
        self.assertEqual(repr(ht2), str_rep)

        self.validate_dict_is_json_compat(ht2.chart_data)

    def test_pca(self):
        """Test PCA"""
        pca = stats.PCA(self.expected_arr_no_nan)
        exp = np.array(
            [
                [
                    0.26641603,
                    0.24490217,
                    0.32525642,
                    0.49217851,
                    0.55671381,
                    0.45943851,
                ],
                [
                    -0.55407909,
                    -0.56054476,
                    -0.38688698,
                    0.19060815,
                    0.33893785,
                    0.2790951,
                ],
            ]
        )
        assert_array_almost_equal(pca.feature_map_data, exp)
        self.assertEqual(pca.component_labels, ["1st Comp", "2nd Comp"])

        pca = stats.PCA(self.expected_arr_no_nan, transform=False)
        assert_array_almost_equal(pca.feature_map_data, exp)

    def test_multi_variable_regression(self):
        """Test Multi-Variable Linear Regression"""
        y = np.linspace(1, 10, 10)
        mvr = stats.MultiVariableRegression(self.expected_arr_no_nan, y)
        self.assertEqual(round(mvr.y_intercept, 3), 2.983)
        slope = np.array(
            [
                3.80907501,
                -4.68155991,
                0.67410364,
                -0.32432575,
                0.9811169,
                -1.10516451,
            ]
        )
        assert_array_almost_equal(mvr.slope, slope)
        self.assertEqual(round(mvr.mse, 3), 1.353)
        self.assertEqual(round(mvr.r_sq, 3), 0.836)
        residuals = np.array(
            [
                -0.62566871,
                -1.28605435,
                0.47511697,
                0.40490283,
                -0.11474329,
                -0.13508221,
                0.34435215,
                -1.26302838,
                -0.76163533,
                2.96184032,
            ]
        )
        assert_array_almost_equal(mvr.residuals, residuals)
        self.assertEqual(round(mvr.f_stat, 3), 2.548)
        self.assertEqual(mvr.df_error, 3)
        self.assertEqual(mvr.df_model, 6)
        self.assertEqual(round(mvr.f_p_value, 3), 0.763)

        mvr2 = stats.MultiVariableRegression(
            self.expected_arr_no_nan, y, saved_reg=mvr
        )
        assert_array_equal(mvr.residuals, mvr2.residuals)

        self.validate_dict_is_json_compat(mvr.chart_data)
        self.validate_dict_is_json_compat(mvr.prob_plot)

    def test_box_cox(self):
        """Test Box-Cox"""
        stats.box_cox(self.expected_arr_no_nan[:, 0])
        self.assertTrue(stats.is_nan_arr(stats.box_cox(np.ones(10))))
        with self.assertRaises(ValueError):
            stats.is_nan_arr(stats.box_cox(np.ones(10), const_policy="raise"))

    def test_moving_avg(self):
        """Test moving average calculation"""
        avg_len = 5
        y = np.array([4, 23, 1, 4, -18, 2.2, 100, 34, 9, 8])
        x = np.linspace(1, len(y), len(y))
        x_avg, y_avg = stats.moving_avg(y, avg_len)
        assert_array_equal(x_avg, x[avg_len - 1 : :])
        y_avg_exp = np.array([2.8, 2.44, 17.84, 24.44, 25.44, 30.64])
        assert_array_almost_equal(y_avg, y_avg_exp)

        avg_len = 3
        x += 10
        w = np.array([1, 2, 1, 1, 5, 2, 1, 1, 1, 1])
        x_avg, y_avg = stats.moving_avg(y, avg_len, x=x, weight=w)
        assert_array_equal(x_avg, x[avg_len - 1 : :])
        y_avg_exp = np.array(
            [5.5, 5.5, 0.466667, 0.5, 32.5, 45.033333, 47.666667, 17.0]
        )
        assert_array_almost_equal(y_avg, y_avg_exp)

    def test_avg_moving_range(self):
        """Test avg moving range"""
        test = np.array([1, 2, 3, 4, np.nan])
        self.assertEqual(stats.avg_moving_range(test), 1.0)
        test_nan = np.array([np.nan])
        self.assertTrue(np.isnan(stats.avg_moving_range(test_nan)))

    def test_is_arr_const(self):
        """Test if array is constant"""
        self.assertTrue(stats.is_arr_constant(np.ones(5)))
        self.assertFalse(stats.is_arr_constant(np.array([1, 2])))

    def test_process_nan_policy(self):
        """Test process nan policy"""
        arr = np.array([1, 2, 3, np.nan])
        test = stats.process_nan_policy(arr, "omit")
        assert_array_equal(test, arr[:3])

        with self.assertRaises(NotImplementedError):
            stats.process_nan_policy(arr, "raise")

        test = stats.process_nan_policy(arr, "propagate")
        self.assertTrue(np.isnan(test))

    def test_histogram(self):
        """Test histogram class object"""
        hist = stats.Histogram(np.linspace(1, 20, 20), 4)
        data = hist.chart_data
        assert_array_equal(data["x"], [3.375, 8.125, 12.875, 17.625])
        assert_array_equal(data["y"], [5, 5, 5, 5])
        self.assertEqual(data["mean"], 10.5)
        self.assertEqual(data["median"], 10.5)
        self.assertEqual(round(data["std"], 3), 5.766)
        self.assertEqual(round(data["normality"], 3), 3.992)
        self.assertEqual(round(data["normality_p"], 3), 0.136)

        self.validate_dict_is_json_compat(hist.chart_data)

    def validate_dict_is_json_compat(self, data_dict):
        """Verify dictionary is json compatible"""
        for key, value in data_dict.items():
            self.assertIsInstance(key, (int, str))
            if isinstance(value, list):
                self.validate_list_is_json_compat(value)
            elif isinstance(value, dict):
                self.validate_dict_is_json_compat(value)
            else:
                self.assertIsInstance(value, (float, str, int))

    def validate_list_is_json_compat(self, data_list):
        """Verify list is json compatible"""
        for item in data_list:
            if isinstance(item, list):
                self.validate_list_is_json_compat(item)
            elif isinstance(item, dict):
                self.validate_dict_is_json_compat(item)
            else:
                self.assertIsInstance(item, (float, str, int))


if __name__ == "__main__":
    import sys

    sys.exit(unittest.main())



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# ui.py
"""DVHA-Stats classes for user interaction"""
#
# Copyright (c) 2020 Dan Cutright
# Copyright (c) 2020 Arka Roy
# This file is part of DVHA-Stats, released under a MIT license.
#    See the file LICENSE included with this distribution, also
#    available at https://github.com/cutright/DVHA-Stats


import numpy as np
from dvhastats.utilities import import_data
from dvhastats import plot
from dvhastats import stats


class DVHAStatsBaseClass:
    """Base Class for DVHAStats objects and child objects"""

    def __init__(self):
        """Initialization of DVHAStatsBaseClass for common attr/methods"""
        self.plots = []

    def close(self, figure_number):
        """Close a plot by figure_number"""
        for i, p in enumerate(self.plots):
            if p.figure.number == figure_number:
                p.close()
                self.plots.pop(i)
                return


class DVHAStats(DVHAStatsBaseClass):
    """The main UI class object for DVHAStats"""

    def __init__(
        self,
        data,
        var_names=None,
        x_axis=None,
        avg_len=5,
        del_const_vars=False,
    ):
        """Class used to calculated various statistics

        Parameters
        ----------
        data : numpy.array, dict, str
            Input data (2-D) with N rows of observations and
            p columns of variables.  The CSV file must have a header row
            for column names.
        var_names : list of str, optional
            If data is a numpy array, optionally provide the column names.
        x_axis : numpy.array, list, optional
            Specify x_axis for plotting purposes. Default is based on row
            number in data
        avg_len : int
            When plotting raw data, a trend line will be plotted using this
            value as an averaging length. If N < avg_len + 1 will not
            plot a trend line
        del_const_vars : bool
            Automatically delete any variables that have constant data. The
            names of these variables are stored in the excluded_vars attr.
            Default value is False.
        """
        DVHAStatsBaseClass.__init__(self)

        self.data, self.var_names = import_data(data, var_names)

        self.x_axis = x_axis

        self.deleted_vars = []

        self.box_cox_data = None

        self.avg_len = avg_len

        if del_const_vars:
            self.del_const_vars()

    def get_data_by_var_name(self, var_name):
        """Get the single variable array based on var_name

        Parameters
        ----------
        var_name : int, str
            The name (str) or index (int) of the variable of interest

        Returns
        ----------
        np.ndarray
            The column of data for the given var_name
        """
        index = self.get_index_by_var_name(var_name)
        return self.data[:, index]

    def get_index_by_var_name(self, var_name):
        """Get the variable index by var_name

        Parameters
        ----------
        var_name : int, str
            The name (str) or index (int) of the variable of interest

        Returns
        ----------
        int
            The column index for the given var_name
        """
        if var_name in self.var_names:
            index = self.var_names.index(var_name)
        elif isinstance(var_name, int) and var_name in range(
            self.variable_count
        ):
            return var_name
        else:
            msg = "%s is not a valid var_name\n%s" % (
                var_name,
                ",".join(self.var_names),
            )
            raise AttributeError(msg)
        return index

    @property
    def observations(self):
        """Number of observations in data

        Returns
        ----------
        int
            Number of rows in data
        """
        return self.data.shape[0]

    @property
    def variable_count(self):
        """Number of variables in data

        Returns
        ----------
        int
            Number of columns in data"""
        return self.data.shape[1]

    def del_const_vars(self):
        """Permanently remove variables with no variation"""
        self.deleted_vars = self.constant_vars
        self.data = self.non_const_data

    def correlation_matrix(self, corr_type="Pearson"):
        """Get a Pearson-R or Spearman correlation matrices

        Parameters
        ----------
        corr_type : str
            Either "Pearson" or "Spearman"

        Returns
        ----------
        CorrelationMatrixUI
            A CorrelationMatrixUI class object
        """
        return CorrelationMatrixUI(
            self.data, self.var_names, corr_type=corr_type
        )

    def is_constant(self, var_name):
        """Determine if data by var_name is constant

        Parameters
        ----------
        var_name : int, str
            The var_name to check (or index of variable)

        Returns
        ----------
        bool
            True if all values of var_name are the same (i.e., no variation)
        """
        data = self.get_data_by_var_name(var_name)
        return stats.is_arr_constant(data)

    @property
    def constant_vars(self):
        """Get a list of all constant variables

        Returns
        ----------
        list
            Names of variables with no variation
        """
        return [v for v in self.var_names if self.is_constant(v)]

    @property
    def constant_var_indices(self):
        """Get a list of all constant variable indices

        Returns
        ----------
        list
            Indices of variables with no variation
        """
        return [i for i, v in enumerate(self.var_names) if self.is_constant(v)]

    @property
    def non_const_data(self):
        """Return self.data excluding any constant variables

        Returns
        ----------
        np.ndarray
            Data with constant variables removed. This does not alter the data
            property.
        """
        return np.delete(self.data, self.constant_var_indices, axis=1)

    def linear_reg(self, y, saved_reg=None):
        """Initialize a MultiVariableRegression class object

        Parameters
        ----------
        y : np.ndarray, list
            Dependent data based on DVHAStats.data
        saved_reg : MultiVariableRegression, optional
            If supplied, predicted values (y-hat) will be calculated with
            DVHAStats.data and the regression from saved_reg. This is useful
            if testing a regression model on new data.

        Returns
        ----------
        LinearRegUI
            A LinearRegUI class object.
        """
        return LinearRegUI(self.data, y, saved_reg)

    def univariate_control_charts(
        self,
        std=3,
        ucl_limit=None,
        lcl_limit=None,
        box_cox=False,
        box_cox_alpha=None,
        box_cox_lmbda=None,
        const_policy="propagate",
    ):
        """
        Calculate control limits for a standard univariate Control Chart

        Parameters
        ----------
        std : int, float, optional
            Number of standard deviations used to calculate if a y-value is
            out-of-control
        ucl_limit : float, optional
            Limit the upper control limit to this value
        lcl_limit : float, optional
            Limit the lower control limit to this value
        box_cox : bool, optional
            Set to true to perform a Box-Cox transformation on data prior to
            calculating the control chart statistics
        box_cox_alpha : float, optional
            If alpha is not None, return the 100 * (1-alpha)% confidence
            interval for lmbda as the third output argument. Must be between
            0.0 and 1.0.
        box_cox_lmbda : float, optional
            If lmbda is not None, do the transformation for that value.
            If lmbda is None, find the lambda that maximizes the log-likelihood
            function and return it as the second output argument.
        const_policy : str
            {‘propagate’, ‘raise’, 'omit'}
            Defines how to handle when data is constant. The following
            options are available (default is ‘propagate’):
            ‘propagate’: returns nan
            ‘raise’: throws an error
            'omit': remove NaN data

        Returns
        ----------
        dict
            ControlChart class objects stored in a dictionary with
            var_names and indices as keys (can use var_name or index)
        """
        kwargs = {"std": std, "ucl_limit": ucl_limit, "lcl_limit": lcl_limit}
        data = {}
        if box_cox:
            if self.box_cox_data is None:
                self.box_cox(
                    alpha=box_cox_alpha,
                    lmbda=box_cox_lmbda,
                    const_policy=const_policy,
                )
            cc_data = self.box_cox_data
            plot_title = "Univariate Control Chart with Box-Cox Transformation"
        else:
            cc_data = self.data
            plot_title = None
        for i, key in enumerate(self.var_names):
            if const_policy == "propagate" and stats.is_nan_arr(cc_data[:, i]):
                plot_title = "Cannot calculate control chart with const data!"
            data[key] = ControlChartUI(
                cc_data[:, i], var_name=key, plot_title=plot_title, **kwargs
            )
            data[i] = data[key]
        return data

    def hotelling_t2(
        self,
        alpha=0.05,
        box_cox=False,
        box_cox_alpha=None,
        box_cox_lmbda=None,
        const_policy="omit",
    ):
        """
        Calculate control limits for a standard univariate Control Chart

        Parameters
        ----------
        alpha : float
            Significance level used to determine the upper control limit (ucl)
        box_cox : bool, optional
            Set to true to perform a Box-Cox transformation on data prior to
            calculating the control chart statistics
        box_cox_alpha : float, optional
            If alpha is not None, return the 100 * (1-alpha)% confidence
            interval for lmbda as the third output argument. Must be between
            0.0 and 1.0.
        box_cox_lmbda : float, optional
            If lmbda is not None, do the transformation for that value.
            If lmbda is None, find the lambda that maximizes the log-likelihood
            function and return it as the second output argument.
        const_policy : str
            {‘raise’, 'omit'}
            Defines how to handle when data is constant. The following
            options are available (default is ‘raise’):
            ‘raise’: throws an error
            'omit': exclude constant variables from calculation

        Returns
        ----------
        HotellingT2UI
            HotellingT2UI class object
        """

        if box_cox:
            if self.box_cox_data is None:
                self.box_cox(
                    alpha=box_cox_alpha,
                    lmbda=box_cox_lmbda,
                    const_policy=const_policy,
                )
            data = self.box_cox_data
            if const_policy == "omit":
                data = stats.remove_const_column(data)
            plot_title = (
                "Multivariate Control Chart with Box-Cox Transformation"
            )
        else:
            data = self.non_const_data if const_policy == "omit" else self.data
            plot_title = None

        return HotellingT2UI(data, alpha, plot_title=plot_title)

    def box_cox_by_index(
        self, index, alpha=None, lmbda=None, const_policy="propagate"
    ):
        """

        Parameters
        ----------
        index : int, str
            The index corresponding to the variable data to have a box-cox
            transformation applied.  If index is a string, it will be assumed
            to be the var_name
        lmbda : None, scalar, optional
            If lmbda is not None, do the transformation for that value.
            If lmbda is None, find the lambda that maximizes the
            log-likelihood function and return it as the second output
            argument.
        alpha : None, float, optional
            If alpha is not None, return the 100 * (1-alpha)% confidence
            interval for lmbda as the third output argument. Must be between
            0.0 and 1.0.
        const_policy : str
            {‘propagate’, ‘raise’, 'omit'}
            Defines how to handle when data is constant. The following
            options are available (default is ‘propagate’):
            ‘propagate’: returns nan
            ‘raise’: throws an error
            'omit': remove

        Returns
        ----------
        np.ndarray
            Results from stats.box_cox
        """
        if self.box_cox_data is None:
            self.box_cox_data = np.zeros_like(self.data)

        if isinstance(index, str):
            index = self.get_index_by_var_name(index)

        self.box_cox_data[:, index] = stats.box_cox(
            self.data[:, index],
            alpha=alpha,
            lmbda=lmbda,
            const_policy=const_policy,
        )

        return self.box_cox_data[:, index]

    def box_cox(self, alpha=None, lmbda=None, const_policy="propagate"):
        """Apply box_cox_by_index for all data"""
        for i in range(self.variable_count):
            self.box_cox_by_index(
                i, alpha=alpha, lmbda=lmbda, const_policy=const_policy
            )

    def pca(self, n_components=0.95, transform=True, **kwargs):
        """Return an sklearn PCA-like object, see PCA object for details

        Parameters
        ----------
        n_components : int, float, None or str
            Number of components to keep. if n_components is not set all
            components are kept:
            n_components == min(n_samples, n_features)

            If n_components == 'mle' and svd_solver == 'full', Minka’s MLE
            is used to guess the dimension. Use of n_components == 'mle'
            will interpret svd_solver == 'auto' as svd_solver == 'full'.

            If 0 < n_components < 1 and svd_solver == 'full', select the
            number of components such that the amount of variance that
            needs to be explained is greater than the percentage specified
            by n_components.

            If svd_solver == 'arpack', the number of components must be
            strictly less than the minimum of n_features and n_samples.
        transform : bool
            Fit the model and apply the dimensionality reduction
        kwargs : any
            Provide any keyword arguments for sklearn.decomposition.PCA:
            https://scikit-learn.org/stable/modules/generated/sklearn.decomposition.PCA.html

        Returns
        ----------
        stats.PCA
            A principal component analysis object inherited from
            sklearn.decomposition.PCA
        """
        return PCAUI(
            self.data,
            var_names=self.var_names,
            n_components=n_components,
            transform=transform,
            **kwargs
        )

    def __add_tend_line(self, var_name, plot_index):
        """Add trend line based on moving average"""
        trend_x, trend_y = stats.moving_avg(
            self.get_data_by_var_name(var_name), self.avg_len
        )
        self.plots[plot_index].add_line(
            trend_y, x=trend_x, line_color="black", line_width=0.75
        )

    def show(self, var_name, plot_type="trend", **kwargs):
        """Display a plot of var_name with matplotlib

        Parameters
        ----------
        var_name : str, int
            The name (str) or index (int) of teh variable to plot
        plot_type : str
            Either "trend" or "hist"
        kwargs : any
            If plot_type is "hist", pass any of the matplotlib hist key word
            arguments

        Returns
        ----------
        int
            The number of the newly created matplotlib figure
        """
        index = self.get_index_by_var_name(var_name)
        var_name = self.var_names[index]
        if plot_type == "trend":
            self.plots.append(
                plot.Plot(
                    self.data[:, index],
                    x=self.x_axis,
                    xlabel="Observation",
                    ylabel=var_name,
                    title="",
                )
            )
            self.__add_tend_line(var_name, -1)
        elif plot_type == "hist":
            self.plots.append(
                plot.Histogram(self.data[:, index], xlabel=var_name, **kwargs)
            )
        return self.plots[-1].figure.number


class ControlChartUI(DVHAStatsBaseClass, stats.ControlChart):
    """Univariate Control Chart"""

    def __init__(
        self,
        y,
        std=3,
        ucl_limit=None,
        lcl_limit=None,
        var_name=None,
        x=None,
        plot_title=None,
    ):
        """
        Calculate control limits for a standard univariate Control Chart

        Parameters
        ----------
        y : list, np.ndarray
            Input data (1-D)
        std : int, float, optional
            Number of standard deviations used to calculate if a y-value is
            out-of-control.
        ucl_limit : float, optional
            Limit the upper control limit to this value
        lcl_limit : float, optional
            Limit the lower control limit to this value
        plot_title : str, optional
            Over-ride the plot title
        """
        DVHAStatsBaseClass.__init__(self)
        stats.ControlChart.__init__(
            self, y, std=std, ucl_limit=ucl_limit, lcl_limit=lcl_limit, x=x
        )

        self.plot_title = (
            "Univariate Control Chart" if plot_title is None else plot_title
        )

        self.var_name = var_name
        self.plots = []

    def show(self):
        """Display the univariate control chart with matplotlib

        Returns
        ----------
        int
            The number of the newly created matplotlib figure
        """
        self.plots.append(
            plot.ControlChart(
                title=self.plot_title, ylabel=self.var_name, **self.chart_data
            )
        )
        return self.plots[-1].figure.number


class HotellingT2UI(DVHAStatsBaseClass, stats.HotellingT2):
    """Hotelling's t-squared statistic for multivariate hypothesis testing"""

    def __init__(self, data, alpha=0.05, plot_title=None):
        """Initialize the Hotelling T^2 class

        Parameters
        ----------
        data : np.ndarray
            A 2-D array of data to perform multivariate analysis.
            (e.g., DVHAStats.data)
        alpha : float
            The significance level used to calculate the
            upper control limit (UCL)
        plot_title : str, optional
            Over-ride the plot title
        """
        DVHAStatsBaseClass.__init__(self)
        stats.HotellingT2.__init__(self, data, alpha=alpha)

        self.plot_title = (
            "Multivariate Control Chart" if plot_title is None else plot_title
        )
        self.plots = []

    def show(self):
        """Display the multivariate control chart with matplotlib

        Returns
        ----------
        int
            The number of the newly created matplotlib figure
        """
        self.plots.append(
            plot.ControlChart(
                title=self.plot_title,
                ylabel="Hottelling T^2",
                **self.chart_data
            )
        )
        return self.plots[-1].figure.number


class PCAUI(DVHAStatsBaseClass, stats.PCA):
    """Hotelling's t-squared statistic for multivariate hypothesis testing"""

    def __init__(
        self, X, var_names=None, n_components=0.95, transform=True, **kwargs
    ):
        """Initialize PCA and perform fit. Inherits sklearn.decomposition.PCA

        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
            Training data, where n_samples is the number of samples and
            n_features is the number of features.
        var_names : str, optional
            Names of the independent variables in X
        n_components : int, float, None or str
            Number of components to keep. if n_components is not set all
            components are kept:
            n_components == min(n_samples, n_features)

            If n_components == 'mle' and svd_solver == 'full', Minka’s MLE
            is used to guess the dimension. Use of n_components == 'mle'
            will interpret svd_solver == 'auto' as svd_solver == 'full'.

            If 0 < n_components < 1 and svd_solver == 'full', select the
            number of components such that the amount of variance that
            needs to be explained is greater than the percentage specified
            by n_components.

            If svd_solver == 'arpack', the number of components must be
            strictly less than the minimum of n_features and n_samples.
        transform : bool
            Fit the model and apply the dimensionality reduction
        kwargs : any
            Provide any keyword arguments for sklearn.decomposition.PCA:
            https://scikit-learn.org/stable/modules/generated/sklearn.decomposition.PCA.html
        """
        # print(kwargs)
        DVHAStatsBaseClass.__init__(self)
        stats.PCA.__init__(
            self, X, n_components=n_components, transform=transform, **kwargs
        )
        self.var_names = range(X.shape[1]) if var_names is None else var_names
        self.plots = []

    def show(self, plot_type="feature_map", absolute=True):
        """Create a heat map of PCA components

        Parameters
        ----------
        plot_type : str
            Select a plot type to display. Options include: feature_map.
        absolute : bool
            Heat map will display the absolute values in PCA components
            if True

        Returns
        ----------
        int
            The number of the newly created matplotlib figure
        """
        if plot_type == "feature_map":
            data = self.feature_map_data
            if absolute:
                data = abs(data)
            self.plots.append(plot.PCAFeatureMap(data, self.var_names))
            return self.plots[-1].figure.number


class CorrelationMatrixUI(DVHAStatsBaseClass, stats.CorrelationMatrix):
    """Pearson-R correlation matrix UI object"""

    def __init__(
        self, X, var_names=None, corr_type="Pearson", cmap="coolwarm"
    ):
        """Initialization of CorrelationMatrix object

        Parameters
        ----------
        X : np.ndarray
            Input data (2-D) with N rows of observations and
            p columns of variables.
        var_names : list, optional
            Optionally set the variable names with a list of str
        corr_type : str
            Either "Pearson" or "Spearman"
        cmap : str
            matplotlib compatible color map
        """
        DVHAStatsBaseClass.__init__(self)
        stats.CorrelationMatrix.__init__(self, X=X, corr_type=corr_type)
        self.var_names = range(X.shape[1]) if var_names is None else var_names
        self.cmap = cmap

    def show(self, absolute=False, corr=True):
        """Create a heat map of PCA components

        Parameters
        ----------
        absolute : bool
            Heat map will display the absolute values in PCA components
            if True
        corr : bool
            Plot a p-value matrix if False, correlation matrix if True.

        Returns
        ----------
        int
            The number of the newly created matplotlib figure
        """

        data = self.corr if corr else self.p
        data = abs(data) if absolute else data

        mat_type = "Pearson-R" if self.corr_type == "pearson" else "Spearman"
        value_type = ["p-value", "Correlation"][corr]
        title = "%s %s Matrix" % (mat_type, value_type)

        self.plots.append(
            plot.HeatMap(
                data,
                xlabels=self.var_names,
                ylabels=self.var_names,
                cmap=self.cmap,
                title=title,
            )
        )
        return self.plots[-1].figure.number


class LinearRegUI(DVHAStatsBaseClass, stats.MultiVariableRegression):
    """A MultiVariableRegression class UI object

    Parameters
    ----------
    y : np.ndarray, list
        Dependent data based on DVHAStats.data
    saved_reg : MultiVariableRegression, optional
        If supplied, predicted values (y-hat) will be calculated with
        DVHAStats.data and the regression from saved_reg. This is useful
        if testing a regression model on new data.
    """

    def __init__(self, X, y, saved_reg=None):
        DVHAStatsBaseClass.__init__(self)
        stats.MultiVariableRegression.__init__(
            self, X=X, y=y, saved_reg=saved_reg
        )

    def show(self, plot_type="residual"):
        """Create a Residual or Probability Plot

        Parameters
        ----------
        plot_type : str
            Either "residual" or "prob"

        Returns
        ----------
        int
            The number of the newly created matplotlib figure
        """

        if plot_type not in {"residual", "prob"}:
            return

        title = (
            "Multi-Variable Linear Regression"
            if self.X.shape[1] > 1
            else "Linear Regression"
        )

        if plot_type == "residual":
            data = self.chart_data
            self.plots.append(
                plot.Plot(
                    data["resid"],
                    x=data["y"],
                    title=title,
                    xlabel="Fitted Values",
                    ylabel="Residual",
                    line=False,
                )
            )
            x_zero = [np.min(data["y"]), np.max(data["y"])]
            y_zero = [0, 0]
            self.plots[-1].add_line(
                y_zero, x_zero, line_color="black", line_style="--"
            )
        elif plot_type == "prob":
            data = self.prob_plot
            self.plots.append(
                plot.Plot(
                    data["y"],
                    x=data["x"],
                    title="Probability Plot",
                    xlabel="Quantiles",
                    ylabel="Ordered Values",
                    line=False,
                )
            )
            self.plots[-1].add_line(
                data["y_trend"],
                data["x_trend"],
                line_color="black",
                line_style="--",
            )

        return self.plots[-1].figure.number



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# plot.py
"""Basic plotting class objects for DVHA-Stats based on matplotlib"""
#
# Copyright (c) 2020 Dan Cutright
# This file is part of DVHA-Stats, released under a MIT license.
#    See the file LICENSE included with this distribution, also
#    available at https://github.com/cutright/DVHA-Stats

from matplotlib import pyplot as plt
import numpy as np


FIGURE_COUNT = 1


def get_new_figure_num():
    """Get a number for a new matplotlib figure

    Returns
    ----------
    int
        Figure number
    """
    global FIGURE_COUNT
    FIGURE_COUNT += 1
    return FIGURE_COUNT - 1


class Chart:
    """Base class for charts"""

    def __init__(self, title=None, fig_init=True):
        """Initialization of Chart base class"""
        self.title = title
        self.figure = plt.figure(get_new_figure_num()) if fig_init else None

        if title and fig_init:
            self.figure.suptitle(title, fontsize=16)

    def show(self):
        """Display this figure"""
        self.activate()
        plt.show()

    def activate(self):
        """Activate this figure"""
        plt.figure(self.figure.number)

    def close(self):
        """Close this figure"""
        plt.close(self.figure.number)


class Plot(Chart):
    """Generic plotting class with matplotlib"""

    def __init__(
        self,
        y,
        x=None,
        show=True,
        title="Chart",
        xlabel="Independent Variable",
        ylabel="Dependent Variable",
        line=True,
        line_color=None,
        line_width=1.0,
        line_style="-",
        scatter=True,
        scatter_color=None,
    ):
        """
        Initialization of a general Plot class object

        Parameters
        ----------
        y : np.ndarray, list
            The y data to be plotted (1-D only)
        x : np.ndarray, list, optional
            Optionally specify the x-axis values. Otherwise index+1 is used.
        show : bool
            Automatically plot the data if True
        title : str
            Set the plot title
        xlabel : str
            Set the x-axis title
        ylabel : str
            Set the y-axis title
        line : bool
            Plot the data as a line series
        line_color : str, optional
            Specify the line color
        line_width : float, int
            Specify the line width
        line_style : str
            Specify the line style
        scatter : bool
            Plot the data as a scatter plot (circles)
        scatter_color : str, optional
            Specify the scatter plot circle color
        """
        Chart.__init__(self, title=title)
        self.x = np.linspace(1, len(y), len(y)) if x is None else np.array(x)
        self.y = np.array(y) if not isinstance(y, np.ndarray) else y
        self.show = show
        self.xlabel = xlabel
        self.ylabel = ylabel

        self.line = line
        self.line_color = line_color
        self.line_width = line_width
        self.line_style = line_style
        self.scatter = scatter
        self.scatter_color = scatter_color
        self.activate()

        self.__add_labels()
        self.__add_data()

        if show:
            plt.show()

    def __add_labels(self):
        """Set the x and y axes labels to figure"""
        plt.xlabel(self.xlabel)
        plt.ylabel(self.ylabel)

    def __add_data(self):
        """Add scatter and/or line data to figure"""
        if self.scatter:
            self.add_scatter()
        if self.line:
            self.add_default_line()

    def add_scatter(self):
        """Add scatter data to figure"""
        self.activate()
        plt.scatter(self.x, self.y, color=self.scatter_color)

    def add_default_line(self):
        """Add line data to figure"""
        self.activate()
        plt.plot(
            self.x,
            self.y,
            color=self.line_color,
            linewidth=self.line_width,
            linestyle=self.line_style,
        )

    def add_line(
        self, y, x=None, line_color=None, line_width=None, line_style=None
    ):
        """Add another line with the provided data

        Parameters
        ----------
        y : np.ndarray, list
            The y data to be plotted (1-D only)
        x: np.ndarray, list, optional
            Optionally specify the x-axis values. Otherwise index+1 is used.
        line_color: str, optional
            Specify the line color
        line_width: float, int
            Specify the line width
        line_style : str
            Specify the line style
        """
        self.activate()
        plt.plot(
            np.linspace(1, len(y), len(y)) if x is None else x,
            y,
            color=line_color,
            linewidth=line_width,
            linestyle=line_style,
        )


class ControlChart(Plot):
    """ControlChart class object"""

    def __init__(
        self,
        y,
        out_of_control,
        center_line,
        lcl=None,
        ucl=None,
        title="Control Chart",
        xlabel="Observation",
        ylabel="Charting Variable",
        line_color="black",
        line_width=0.75,
        center_line_color="black",
        center_line_width=1.0,
        center_line_style="--",
        limit_line_color="red",
        limit_line_width=1.0,
        limit_line_style="--",
        **kwargs
    ):
        """Initialization of a ControlChart plot class object

        Parameters
        ----------
        y : np.ndarray, list
            Charting data
        out_of_control : np.ndarray, list
            The indices of y that are out-of-control
        center_line : float, np.ndarray
            The center line value (e.g., np.mean(y))
        lcl : float, optional
            The lower control limit (LCL). Line omitted if lcl is None.
        ucl : float, optional
            The upper control limit (UCL). Line omitted if ucl is None.
        title: str
            Set the plot title
        xlabel: str
            Set the x-axis title
        ylabel: str
            Set the y-axis title
        line_color: str, optional
            Specify the line color
        line_width: float, int
            Specify the line width
        kwargs : any
            Any additional keyword arguments applicable to the Plot class
        """
        self.center_line = center_line
        self.lcl = lcl
        self.ucl = ucl
        self.out_of_control = out_of_control
        self.center_line_color = center_line_color
        self.center_line_width = center_line_width
        self.center_line_style = center_line_style
        self.limit_line_color = limit_line_color
        self.limit_line_width = limit_line_width
        self.limit_line_style = limit_line_style
        kwargs["title"] = title
        kwargs["xlabel"] = xlabel
        kwargs["ylabel"] = ylabel
        kwargs["line_color"] = line_color
        kwargs["line_width"] = line_width
        Plot.__init__(self, y, **kwargs)
        self.__add_cc_data()
        self.__add_table_with_limits()

        if self.show:
            plt.show()

    def __set_y_scatter_data(self):
        """Add circles colored by out-of-control status"""
        include = np.full(len(self.y), True)
        for i in self.out_of_control:
            include[i] = False
        self.ic = {"x": self.x[include], "y": self.y[include]}
        self.ooc = {"x": self.x[~include], "y": self.y[~include]}

    def __add_cc_data(self):
        """Add center line and upper/lower control limit lines"""
        self.add_control_limit_line(self.ucl)
        self.add_control_limit_line(self.lcl)
        self.add_center_line()

    def __add_table_with_limits(self):
        """Add tables with center line and upper/lower control limit values"""
        self.activate()
        plt.subplots_adjust(bottom=0.25)
        plt.table(
            cellText=self.__table_text,
            cellLoc="center",
            colLabels=["Center Line", "LCL", "UCL"],
            loc="bottom",
            bbox=[0.0, -0.31, 1, 0.12],
        )

    @property
    def __table_text(self):
        """Get text to pass into matplotlib table creation"""
        props = ["center_line", "lcl", "ucl"]
        text = []
        for prop in props:
            value = getattr(self, prop)
            if isinstance(value, float):
                formatter = ["E", "f"][9999 > float(value) > 0.1]
                text.append(("%%0.3%s" % formatter) % value)
            else:
                text.append(str(value))
        return [text]

    def add_scatter(self):
        """Set scatter data, add in- and out-of-control circles"""
        self.activate()
        self.__set_y_scatter_data()
        plt.scatter(self.ic["x"], self.ic["y"], color=self.scatter_color)
        plt.scatter(self.ooc["x"], self.ooc["y"], color="red")

    def add_control_limit_line(
        self, limit, color=None, line_width=None, line_style=None
    ):
        """Add a control limit line to plot"""
        self.activate()
        color = self.limit_line_color if color is None else color
        line_width = (
            self.limit_line_width if line_width is None else line_width
        )
        line_style = (
            self.limit_line_style if line_style is None else line_style
        )
        if limit is not None:
            plt.plot(
                [1, len(self.x)],
                [limit] * 2,
                color=color,
                linewidth=line_width,
                linestyle=line_style,
            )

    def add_center_line(self, color=None, line_width=None, line_style=None):
        """Add the center line to the plot"""
        self.activate()
        color = self.center_line_color if color is None else color
        line_width = (
            self.center_line_width if line_width is None else line_width
        )
        line_style = (
            self.center_line_style if line_style is None else line_style
        )
        plt.plot(
            [1, len(self.x)],
            [self.center_line] * 2,
            color=color,
            linewidth=line_width,
            linestyle=line_style,
        )


class HeatMap(Chart):
    """Create a heat map using matplotlib.pyplot.matshow"""

    def __init__(
        self,
        X,
        xlabels=None,
        ylabels=None,
        title=None,
        cmap="viridis",
        show=True,
    ):
        """Initialization of a HeatMap Chart object"""
        Chart.__init__(self, title=title)
        self.X = X
        self.x_labels = range(X.shape[1]) if xlabels is None else xlabels
        self.y_labels = range(X.shape[0]) if ylabels is None else ylabels

        plt.matshow(X, cmap=cmap, fignum=self.figure.number)
        plt.colorbar()
        self.__set_ticks()

        if show:
            self.show()

    def __set_ticks(self):
        """Set tick labels based on x and y labels"""
        plt.xticks(
            range(self.X.shape[1]), self.x_labels, rotation=30, ha="left"
        )
        plt.yticks(range(self.X.shape[0]), self.y_labels, rotation=30)


class PCAFeatureMap(HeatMap):
    """Specialized Heat Map for PCA feature evaluation"""

    def __init__(
        self,
        X,
        features=None,
        cmap="viridis",
        show=True,
        title="PCA Feature Heat Map",
    ):
        """Initialization of a HeatMap Chart object"""
        HeatMap.__init__(
            self,
            X,
            xlabels=features,
            ylabels=self.get_comp_labels(X.shape[0]),
            cmap=cmap,
            show=show,
            title=title,
        )

    def get_comp_labels(self, n_components):
        """Get ylabels for HeatMap"""
        return [
            "%s Comp" % (self.get_ordinal(n + 1)) for n in range(n_components)
        ]

    @staticmethod
    def get_ordinal(n):
        """Convert number to its ordinal (e.g., 1 to 1st)

        Parameters
        ----------
        n : int
            Number to be converted to ordinal

        Returns
        ----------
        str
            the ordinal of n
        """
        return "%d%s" % (
            n,
            "tsnrhtdd"[(n // 10 % 10 != 1) * (n % 10 < 4) * n % 10 :: 4],
        )


class Histogram(Chart):
    """Histogram plotting class object"""

    def __init__(
        self,
        data,
        bins=10,
        title="Histogram",
        xlabel="Bins",
        ylabel="Counts",
        **kwargs
    ):
        """Initialization of Histogram class

        Parameters
        ----------
        data : array-like
            Input array (1-D)
        bins : int, sequence, str
            default: rcParams["hist.bins"] (default: 10)
            If bins is an integer, it defines the number of equal-width
            bins in the range.

            If bins is a sequence, it defines the bin edges, including the
            left edge of the first bin and the right edge of the last bin;
            in this case, bins may be unequally spaced. All but the last
            (righthand-most) bin is half-open. In other words, if bins is:
            [1, 2, 3, 4]
            then the first bin is [1, 2) (including 1, but excluding 2) and
            the second [2, 3). The last bin, however, is [3, 4], which
            includes 4.

            If bins is a string, it is one of the binning strategies supported
            by numpy.histogram_bin_edges: 'auto', 'fd', 'doane', 'scott',
            'stone', 'rice', 'sturges', or 'sqrt'.
        kwargs : any
            Any keyword argument may be set per matplotlib histogram:
            https://matplotlib.org/3.3.1/api/_as_gen/matplotlib.pyplot.hist.html

        """
        self.title = title
        Chart.__init__(self, title=self.title, fig_init=False)

        self.data = data
        self.bins = bins
        self.xlabel = xlabel
        self.ylabel = ylabel
        self.hist_kwargs = kwargs

        self.__set_hist_data()
        self.__set_title()
        self.__add_labels()

    def __set_hist_data(self):
        """Generate histogram data and add to figure"""
        self.figure, self.axes = plt.subplots()
        self.axes.hist(self.data, bins=self.bins, **self.hist_kwargs)

    def __set_title(self):
        """Set the figure title"""
        self.figure.suptitle(self.title, fontsize=16)

    def __add_labels(self):
        """Set the x and y axes labels to figure"""
        plt.xlabel(self.xlabel)
        plt.ylabel(self.ylabel)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# test_plot.py
"""unittest cases for plot."""
#
# Copyright (c) 2020 Dan Cutright
# This file is part of DVHA-Stats, released under a MIT license.
#    See the file LICENSE included with this distribution, also
#    available at https://github.com/cutright/DVHA-Stats


import unittest
from dvhastats import plot
from numpy import random
import matplotlib

matplotlib.use("Template")


class TestPlot(unittest.TestCase):
    """Unit tests for Utilities."""

    def setUp(self):
        """Setup files and base data for utility testing."""
        # self.x = [0, 1]
        self.y = random.standard_normal(10)

    def test_plot_init(self):
        """Test Plot initialization"""
        p = plot.Plot(self.y)
        p.close()

    def test_plot_add_line(self):
        """Test add_line and show=False"""
        p = plot.Plot(self.y, show=False)
        p.add_line([0, 1])
        p.close()

    def test_control_chart(self):
        """Test ControlChart initialization"""
        p = plot.ControlChart(
            self.y,
            random.randint(5, size=3),
            0,
            lcl=-1,
            ucl=1,
        )
        p.close()

    def test_histogram(self):
        """Test Histogram initialization"""
        p = plot.Histogram(self.y)
        p.close()


if __name__ == "__main__":
    import sys

    sys.exit(unittest.main())



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# stats.py
"""Statistical calculations and class objects"""
#
# Copyright (c) 2020 Dan Cutright
# Copyright (c) 2020 Arka Roy
# This file is part of DVHA-Stats, released under a MIT license.
#    See the file LICENSE included with this distribution, also
#    available at https://github.com/cutright/DVHA-Stats


import numpy as np
from scipy import stats as scipy_stats
from sklearn import linear_model
from sklearn.metrics import mean_squared_error, r2_score
from sklearn.decomposition import PCA as sklearn_PCA
from regressors import stats as regressors_stats


class Histogram:
    """Basic histogram plot using matplotlib histogram calculation"""

    def __init__(self, y, bins, nan_policy="omit"):
        """
        Initialization of Histogram class object

        y: array-like
            Input array.
        nan_policy : str
            Value must be one of the following: ‘propagate’, ‘raise’, ‘omit’
            Defines how to handle when input contains nan. The following
            options are available (default is ‘omit’):
            ‘propagate’: returns nan
            ‘raise’: throws an error
            ‘omit’: performs the calculations ignoring nan values
        """

        self.y = process_nan_policy(y, nan_policy)
        self.bins = bins
        self.nan_policy = nan_policy

    @property
    def mean(self):
        """The mean value of the input array

        Returns
        ----------
        np.ndarray
            Mean value of y with np.mean()
        """
        return np.mean(self.y)

    @property
    def median(self):
        """The median value of the input array

        Returns
        ----------
        np.ndarray
            Median value of y with np.median()
        """
        return np.median(self.y)

    @property
    def std(self):
        """The standard deviation of the input array

        Returns
        ----------
        np.ndarray
            Standard deviation of y with np.std()
        """
        return np.std(self.y)

    @property
    def normality(self):
        """The normality and normality p-value of the input array

        Returns
        ----------
        statistic : float
            Normality calculated with scipy.stats.normaltest
        p-value : float
            A 2-sided chi squared probability for the hypothesis test.
        """
        return scipy_stats.normaltest(self.y, nan_policy=self.nan_policy)

    @property
    def chart_data(self):
        """JSON compatible dict for chart generation

        Returns
        ----------
        dict
            Data used for Histogram visuals. Keys include 'x', 'y', 'mean',
            'median', 'std', 'normality', 'normality_p'
        """
        hist, bins = np.histogram(self.y, bins=self.bins)
        center = (bins[:-1] + bins[1:]) / 2.0
        norm, norm_p = self.normality
        return {
            "x": center.tolist(),
            "y": hist.tolist(),
            "mean": float(self.mean),
            "median": float(self.median),
            "std": float(self.std),
            "normality": float(norm),
            "normality_p": float(norm_p),
        }


class MultiVariableRegression:
    """Multi-variable regression using scikit-learn"""

    def __init__(self, X, y, saved_reg=None):
        """Initialization of a MultiVariableRegression

        Parameters
        ----------
        X : array-like
            Independent data
        y : array-like
            Dependent data
        saved_reg : MultiVariableRegression, optional
            Optionally provide a previously calculated regression
        """

        self.X = np.array(X)
        self.y = np.array(y)

        self._do_fit(saved_reg=saved_reg)

    def _do_fit(self, saved_reg=None):
        """Perform linear regression

        Parameters
        ----------
        saved_reg : MultiVariableRegression, optional
            Optionally provide a previously calculated regression
        """
        if saved_reg is None:
            self.reg = linear_model.LinearRegression()
            self.ols = self.reg.fit(self.X, self.y)
        else:
            self.reg = saved_reg.reg
            self.ols = saved_reg.ols

        self.predictions = self.reg.predict(self.X)

        self.p, self.std_err, self.t = get_lin_reg_p_values(
            self.X, self.y, self.predictions, self.y_intercept, self.slope
        )

    @property
    def y_intercept(self):
        """The y-intercept of the linear regression

        Returns
        ----------
        float
            Independent term in the linear model.
        """
        return self.reg.intercept_ if self.reg is not None else None

    @property
    def slope(self):
        """The slope of the linear regression

        Returns
        ----------
        np.ndarray
            Estimated coefficients for the linear regression problem. If
            multiple targets are passed during the fit (y 2D), this is a 2D
            array of shape (n_targets, n_features), while if only one target
            is passed, this is a 1D array of length n_features.
        """
        return self.reg.coef_ if self.reg is not None else None

    @property
    def r_sq(self):
        """R^2 (coefficient of determination) regression score function.

        Returns
        ----------
        float
            The R^2 score
        """
        return r2_score(self.y, self.predictions)

    @property
    def mse(self):
        """Mean squared error of the linear regression

        Returns
        ----------
        float, nd.array
            A non-negative floating point value (the best value is 0.0), or
            an array of floating point values, one for each individual target.
        """
        return mean_squared_error(self.y, self.predictions)

    @property
    def residuals(self):
        """Residuals of the prediction and sample data

        Returns
        ----------
        np.ndarray
            y - predictions
        """
        return np.subtract(self.y, self.predictions)

    @property
    def f_stat(self):
        """The F-statistic of the regression

        Returns
        ----------
        float
            F-statistic of beta coefficients using regressors.stats
        """
        return regressors_stats.f_stat(self.ols, self.X, self.y)

    @property
    def df_error(self):
        """Error degrees of freedom

        Returns
        ----------
        int
            Degrees of freedom for the error
        """
        return len(self.X[:, 0]) - len(self.X[0, :]) - 1

    @property
    def df_model(self):
        """Model degrees of freedom

        Returns
        ----------
        int
            Degrees of freedom for the model
        """
        return len(self.X[0, :])

    @property
    def f_p_value(self):
        """p-value of the f-statistic

        Returns
        ----------
        float
            p-value of the F-statistic of beta coefficients using scipy
        """
        return scipy_stats.f.cdf(self.f_stat, self.df_model, self.df_error)

    @property
    def coef(self):
        """Coefficients for the regression

        Returns
        ----------
        np.ndarray
            An array of regression coefficients (i.e., y_intercept,
            1st var slope, 2nd var slope, etc.)
        """
        return np.concatenate(([self.y_intercept], self.slope))

    @property
    def chart_data(self):
        """JSON compatible dict for chart generation

        Returns
        ----------
        dict
            Data used for residual visuals. Keys include 'x', 'y', 'pred',
            'resid', 'coef', 'r_sq', 'mse', 'std_err', 't_value', 'p_value'
        """
        return {
            "y": self.y.tolist(),
            "pred": self.predictions.tolist(),
            "resid": self.residuals.tolist(),
            "coef": self.coef.tolist(),
            "r_sq": float(self.r_sq),
            "mse": float(self.mse),
            "std_err": self.std_err.tolist(),
            "t_value": self.t.tolist(),
            "p_value": self.p.tolist(),
        }

    @property
    def prob_plot(self):
        """
        Calculate quantiles for a probability plot

        Returns
        ----------
        dict
            Data for generating a probablity plot. Keys include:
            'x', 'y', 'y_intercept', 'slope', 'x_trend', 'y_trend'
        """
        norm_prob_plot = scipy_stats.probplot(
            self.residuals, dist="norm", fit=False, plot=None, rvalue=False
        )

        reg_prob = linear_model.LinearRegression()
        reg_prob.fit([[val] for val in norm_prob_plot[0]], norm_prob_plot[1])

        x_trend = [
            float(np.min(norm_prob_plot[0])),
            float(np.max(norm_prob_plot[0])),
        ]
        y_trend = np.add(
            np.multiply(x_trend, reg_prob.coef_),
            reg_prob.intercept_,
        )
        return {
            "x": norm_prob_plot[0].tolist(),
            "y": norm_prob_plot[1].tolist(),
            "y_intercept": float(reg_prob.intercept_),
            "slope": float(reg_prob.coef_),
            "x_trend": x_trend,
            "y_trend": y_trend.tolist(),
        }


class ControlChart:
    """Univariate Control Chart"""

    def __init__(
        self,
        y,
        std=3,
        ucl_limit=None,
        lcl_limit=None,
        x=None,
    ):
        """
        Calculate control limits for a standard univariate Control Chart

        Parameters
        ----------
        y : list, np.ndarray
            Input data (1-D)
        std : int, float, optional
            Number of standard deviations used to calculate if a y-value is
            out-of-control.
        ucl_limit : float, optional
            Limit the upper control limit to this value
        lcl_limit : float, optional
            Limit the lower control limit to this value
        """

        self.y = np.array(y) if isinstance(y, list) else y
        self.x = x if x else np.linspace(1, len(self.y), len(self.y))
        self.std = std
        self.ucl_limit = ucl_limit
        self.lcl_limit = lcl_limit

        # since moving range is calculated based on 2 consecutive points
        self.scalar_d = 1.128

    def __str__(self):
        """String representation of ControlChartData object"""
        msg = [
            "center_line: %0.3f" % self.center_line,
            "control_limits: %0.3f, %0.3f" % self.control_limits,
            "out_of_control: %s" % self.out_of_control,
        ]
        return "\n".join(msg)

    def __repr__(self):
        """Return the string representation"""
        return str(self)

    @property
    def center_line(self):
        """Center line of charting data (i.e., mean value)

        Returns
        ----------
        np.ndarray, np.nan
            Mean value of y with np.mean() or np.nan if y is empty
        """
        data = remove_nan(self.y)
        if len(data):
            return np.mean(data)
        return np.nan

    @property
    def avg_moving_range(self):
        """Avg moving range based on 2 consecutive points

        Returns
        ----------
        np.ndarray, np.nan
            Average moving range. Returns NaN if arr is empty.
        """
        return avg_moving_range(self.y, nan_policy="omit")

    @property
    def sigma(self):
        """UCL/LCL = center_line +/- sigma * std

        Returns
        ----------
        np.ndarray, np.nan
            sigma or np.nan if arr is empty
        """
        return self.avg_moving_range / self.scalar_d

    @property
    def control_limits(self):
        """Calculate the lower and upper control limits

        Returns
        ----------
        lcl : float
            Lower Control Limit (LCL)
        ucl : float
            Upper Control Limit (UCL)
        """
        cl = self.center_line
        sigma = self.sigma

        ucl = cl + self.std * sigma
        lcl = cl - self.std * sigma

        if self.ucl_limit is not None and ucl > self.ucl_limit:
            ucl = self.ucl_limit
        if self.lcl_limit is not None and lcl < self.lcl_limit:
            lcl = self.lcl_limit

        return lcl, ucl

    @property
    def out_of_control(self):
        """Get the indices of out-of-control observations

        Returns
        ----------
        np.ndarray
            An array of indices that are not between the lower and upper
            control limits
        """
        lcl, ucl = self.control_limits
        high = np.argwhere(self.y > ucl)
        low = np.argwhere(self.y < lcl)
        return np.unique(np.concatenate([high, low]))

    @property
    def out_of_control_high(self):
        """Get the indices of observations > ucl

        Returns
        ----------
        np.ndarray
            An array of indices that are greater than the upper control limit
        """
        _, ucl = self.control_limits
        return np.argwhere(self.y > ucl)

    @property
    def out_of_control_low(self):
        """Get the indices of observations < lcl

        Returns
        ----------
        np.ndarray
            An array of indices that are less than the lower control limit
        """
        lcl, _ = self.control_limits
        return np.argwhere(self.y < lcl)

    @property
    def chart_data(self):
        """JSON compatible dict for chart generation

        Returns
        ----------
        dict
            Data used for Histogram visuals. Keys include 'x', 'y',
            'out_of_control', 'center_line', 'lcl', 'ucl'
        """
        lcl, ucl = self.control_limits
        return {
            "x": self.x.tolist(),
            "y": self.y.tolist(),
            "out_of_control": self.out_of_control.tolist(),
            "center_line": float(self.center_line),
            "lcl": float(lcl),
            "ucl": float(ucl),
        }


class HotellingT2:
    """Hotelling's t-squared statistic for multivariate hypothesis testing"""

    def __init__(self, data, alpha=0.05, const_policy="raise"):
        """Initialize the Hotelling T^2 class

        Parameters
        ----------
        data : np.ndarray
            A 2-D array of data to perform multivariate analysis.
            (e.g., DVHAStats.data)
        alpha : float
            The significance level used to calculate the
            upper control limit (UCL)
        const_policy : str
            {‘raise’, 'omit'}
            Defines how to handle when data is constant. The following
            options are available (default is ‘raise’):
            ‘raise’: throws an error
            'omit': exclude constant variables from calculation
        """

        self.data = (
            data if const_policy == "raise" else remove_const_column(data)
        )
        self.alpha = alpha
        self.lcl = 0

    def __str__(self):
        """String representation of HotellingT2 object"""
        msg = [
            "Q: %s" % self.Q,
            "center_line: %0.3f" % self.center_line,
            "control_limits: %d, %0.3f" % (self.lcl, self.ucl),
            "out_of_control: %s" % self.out_of_control,
        ]
        return "\n".join(msg)

    def __repr__(self):
        """Return the string representation"""
        return str(self)

    @property
    def observations(self):
        """Number of observations in data

        Returns
        ----------
        int
            Number of rows in data
        """
        return self.data.shape[0]

    @property
    def variable_count(self):
        """Number of variables in data

        Returns
        ----------
        int
            Number of columns in data"""
        return self.data.shape[1]

    @property
    def Q(self):
        """Calculate Hotelling T^2 statistic (Q) from a 2-D numpy array

        Returns
        -------
        np.ndarray
            A numpy array of Hotelling T^2 (1-D of length N)
        """
        Q = np.zeros(np.size(self.data, 0))
        D_bar = np.mean(self.data, axis=0)
        S = np.cov(self.data.T)
        S_inv = np.linalg.inv(S)
        observations = np.size(self.data, 0)
        for i in range(observations):
            spread = self.data[i, :] - D_bar
            Q[i] = np.matmul(np.matmul(spread, S_inv), spread)
        return Q

    @property
    def center_line(self):
        """Center line for the control chart

        Returns
        ----------
        float
            Median value of beta distribution.
        """
        return self.get_control_limit(0.5)

    @property
    def control_limits(self):
        """Lower and Upper control limits

        Returns
        ----------
        lcl : float
            Lower Control Limit (LCL). This is fixed to 0 for Hotelling T2
        ucl : float
            Upper Control Limit (UCL)
        """
        return self.lcl, self.ucl

    @property
    def ucl(self):
        """Upper control limit

        Returns
        ----------
        ucl : float
            Upper Control Limit (UCL)"""
        return self.get_control_limit(1 - self.alpha / 2)

    @property
    def out_of_control(self):
        """Indices of out-of-control observations

        Returns
        ----------
        np.ndarray
            An array of indices that are greater than the upper
            control limit. (NOTE: Q is never negative)
        """
        return np.argwhere(self.Q > self.ucl).T[0]

    def get_control_limit(self, x):
        """Calculate a Hotelling T^2 control limit using a beta distribution

        Parameters
        ----------
        x : float
            Value where the beta function is evaluated

        Returns
        -------
        float
            The control limit for a beta distribution
        """

        N = self.observations
        a = self.variable_count / 2
        b = (N - self.variable_count - 1) / 2
        return ((N - 1) ** 2 / N) * scipy_stats.beta.ppf(x, a, b)

    @property
    def chart_data(self):
        """JSON compatible dict for chart generation

        Returns
        ----------
        dict
            Data used for Histogram visuals. Keys include 'x', 'y',
            'out_of_control', 'center_line', 'lcl', 'ucl'
        """
        return {
            "x": list(range(1, self.observations + 1)),
            "y": self.Q.tolist(),
            "out_of_control": self.out_of_control.tolist(),
            "center_line": float(self.center_line),
            "lcl": float(self.lcl),
            "ucl": float(self.ucl),
        }


class PCA(sklearn_PCA):
    """Hotelling's t-squared statistic for multivariate hypothesis testing"""

    def __init__(
        self, X, var_names=None, n_components=0.95, transform=True, **kwargs
    ):
        """Initialize PCA and perform fit. Inherits sklearn.decomposition.PCA

        Parameters
        ----------
        X : np.ndarray
            Training data (2-D), where n_samples is the number of samples and
            n_features is the number of features. shape (n_samples, n_features)
        var_names : list, optional
            Optionally provide names of the features
        n_components : int, float, None or str
            Number of components to keep. if n_components is not set all
            components are kept:
            n_components == min(n_samples, n_features)

            If n_components == 'mle' and svd_solver == 'full', Minka’s MLE
            is used to guess the dimension. Use of n_components == 'mle'
            will interpret svd_solver == 'auto' as svd_solver == 'full'.

            If 0 < n_components < 1 and svd_solver == 'full', select the
            number of components such that the amount of variance that
            needs to be explained is greater than the percentage specified
            by n_components.

            If svd_solver == 'arpack', the number of components must be
            strictly less than the minimum of n_features and n_samples.
        transform : bool
            Fit the model and apply the dimensionality reduction
        kwargs : any
            Provide any keyword arguments for sklearn.decomposition.PCA:
            https://scikit-learn.org/stable/modules/generated/sklearn.decomposition.PCA.html
        """

        self.X = X
        self.var_names = (
            list(range(self.X.shape[1])) if var_names is None else var_names
        )
        sklearn_PCA.__init__(self, n_components=n_components, **kwargs)

        if transform:
            self.fit_transform(self.X)
        else:
            self.fit(self.X)

    @property
    def feature_map_data(self):
        """Used for feature analysis heat map

        Returns
        ----------
        np.ndarray Shape (n_components, n_features)
            Principal axes in feature space, representing the directions of
            maximum variance in the data. The components are sorted by
            explained_variance.
        """
        return self.components_

    @property
    def component_labels(self):
        """Get component names

        Returns
        ----------
        list
            Labels for plotting. (1st Comp, 2nd Comp, 3rd Comp, etc.)
        """
        return [
            "%s Comp" % (get_ordinal(n + 1))
            for n in range(self.components_.shape[0])
        ]


class CorrelationMatrix:
    """Pearson-R correlation matrix"""

    def __init__(self, X, corr_type="Pearson"):
        """Initialization of CorrelationMatrix object

        Parameters
        ----------
        X : np.ndarray
            Input data (2-D) with N rows of observations and
            p columns of variables.
        corr_type : str
            Either "Pearson" or "Spearman"
        """
        self.X = X
        self.corr_type = corr_type.lower()

        if self.corr_type not in {"pearson", "spearman"}:
            msg = "Invalid corr_type: must be either 'Pearson' or 'Spearman'"
            raise NotImplementedError(msg)

        func_map = {
            "pearson": pearson_correlation_matrix,
            "spearman": spearman_correlation_matrix,
        }
        if self.corr_type in list(func_map):
            self.corr, self.p = func_map[self.corr_type](self.X)

    @property
    def normality(self):
        """The normality and normality p-value of the input array

        Returns
        ----------
        statistic : np.ndarray
            Normality calculated with scipy.stats.normaltest
        p-value : np.ndarray
            A 2-sided chi squared probability for the hypothesis test.
        """
        return scipy_stats.normaltest(self.X, nan_policy="omit")

    @property
    def chart_data(self):
        """JSON compatible dict for chart generation

        Returns
        ----------
        dict
            Data used for Histogram visuals. Keys include 'corr', 'p',
            'norm', 'norm_p'
        """
        norm, norm_p = self.normality
        return {
            "corr": self.corr.tolist(),
            "p": self.p.tolist(),
            "norm": norm.tolist(),
            "norm_p": norm_p.tolist(),
        }


#########################################################
# Pure stats functions
#########################################################


def get_lin_reg_p_values(X, y, predictions, y_intercept, slope):
    """
    Get p-values of a linear regression using sklearn
    based on https://stackoverflow.com/questions/27928275/find-p-value-significance-in-scikit-learn-linearregression

    Parameters
    ----------
    X : np.ndarray
        Independent data
    y : np.ndarray, list
        Dependent data
    predictions : np.ndarray, list
        Predictions using the linear regression.
        (Output from linear_model.LinearRegression.predict)
    y_intercept : float, np.ndarray
        The y-intercept of the linear regression
    slope : float, np.ndarray
        The slope of the linear regression

    Returns
    ----------
    p_value : np.ndarray
        p-value of the linear regression coefficients
    std_errs : np.ndarray
        standard errors of the linear regression coefficients
    t_value : np.ndarray
        t-values of the linear regression coefficients
    """

    newX = np.append(np.ones((len(X), 1)), X, axis=1)
    mse = (sum((y - predictions) ** 2)) / (len(newX) - len(newX[0]))

    variance = mse * (np.linalg.inv(np.dot(newX.T, newX)).diagonal())
    std_errs = np.sqrt(variance)

    params = np.append(y_intercept, slope)
    t_values = np.array(params) / std_errs

    p_value = np.array(
        [
            2 * (1 - scipy_stats.t.cdf(np.abs(i), (len(newX) - 1)))
            for i in t_values
        ]
    )

    return p_value, std_errs, t_values


def pearson_correlation_matrix(X):
    """Calculate a correlation matrix of Pearson-R values

    Parameters
    ----------
    X : array-like, shape (n_samples, n_features)
            Training data, where n_samples is the number of samples and
            n_features is the number of features.

    Returns
    ----------
    r : np.ndarray
        Array (2-D) of Pearson-R correlations between the row indexed and
        column indexed variables
    p : np.ndarray
        Array (2-D) of p-values associated with r
    """

    r = np.ones([X.shape[1], X.shape[1]])  # Pearson R
    p = np.zeros([X.shape[1], X.shape[1]])  # p-value of Pearson R

    for x in range(X.shape[1]):
        for y in range(X.shape[1]):
            if x > y:

                # Pearson r requires that both sets of data be of the same
                # length. Remove index if NaN in either variable.
                valid_x = ~np.isnan(X[:, x])
                valid_y = ~np.isnan(X[:, y])
                include = np.full(len(X[:, x]), True)
                for i in range(len(valid_x)):
                    include[i] = valid_x[i] and valid_y[i]
                x_data = X[include, x]
                y_data = X[include, y]

                r[x, y], p[x, y] = scipy_stats.pearsonr(x_data, y_data)

                # These matrices are symmetric
                r[y, x] = r[x, y]
                p[y, x] = p[x, y]

    return r, p


def spearman_correlation_matrix(X, nan_policy="omit"):
    """Calculate a Spearman correlation matrix

    Parameters
    ----------
    X : array-like, shape (n_samples, n_features)
            Training data, where n_samples is the number of samples and
            n_features is the number of features.
    nan_policy : str
        Value must be one of the following: ‘propagate’, ‘raise’, ‘omit’
        Defines how to handle when input contains nan. The following options
        are available (default is ‘omit’):
        ‘propagate’: returns nan
        ‘raise’: throws an error
        ‘omit’: performs the calculations ignoring nan values

    Returns
    ----------
    correlation : float or ndarray (2-D square)
        Spearman correlation matrix or correlation coefficient (if only 2
        variables are given as parameters. Correlation matrix is square with
        length equal to total number of variables (columns or rows) in a and
        b combined.
    p-value : float
        The two-sided p-value for a hypothesis test whose null hypothesis is
        that two sets of data are uncorrelated, has same dimension as rho.

    """
    return scipy_stats.spearmanr(X, nan_policy=nan_policy)


def moving_avg(y, avg_len, x=None, weight=None):
    """Calculate the moving (rolling) average of a set of data

    Parameters
    ----------
    y : array-like
        data (1-D) to be averaged
    avg_len : int
        Data is averaged over this many points (current value and avg_len - 1
        prior points)
    x : np.ndarray, list, optional
        Optionally specify the x-axis values. Otherwise index+1 is used.
    weight : np.ndarray, list, optional
        A weighted moving average is calculated based on the provided weights.
        weight must be of same length as y. Weights of one are assumed by
        default.

    Returns
    ----------
    x : np.ndarray
        Resulting x-values for the moving average
    moving_avg : np.ndarray
        moving average values
    """
    x = np.linspace(1, len(y), len(y)) if x is None else x
    weight = np.ones_like(y) if weight is None else weight

    cumsum, moving_aves, x_final = [0], [], []

    for i, yi in enumerate(y, 1):
        cumsum.append(cumsum[i - 1] + yi / weight[i - 1])
        if i >= avg_len:
            moving_ave = (cumsum[i] - cumsum[i - avg_len]) / avg_len
            moving_aves.append(moving_ave)
    x_final = [x[i] for i in range(avg_len - 1, len(x))]

    return np.array(x_final), np.array(moving_aves)


def box_cox(arr, alpha=None, lmbda=None, const_policy="propagate"):
    """

    Parameters
    ----------
    arr : np.ndarray
        Input array. Must be positive 1-dimensional.
    lmbda : None, scalar, optional
        If lmbda is not None, do the transformation for that value.
        If lmbda is None, find the lambda that maximizes the
        log-likelihood function and return it as the second output
        argument.
    alpha : None, float, optional
        If alpha is not None, return the 100 * (1-alpha)% confidence
        interval for lmbda as the third output argument. Must be between
        0.0 and 1.0.
    const_policy : str
        {‘propagate’, ‘raise’, 'omit'}
        Defines how to handle when data is constant. The following
        options are available (default is ‘propagate’):
        ‘propagate’: returns nan
        ‘raise’: throws an error

    Returns
    ----------
    box_cox : np.ndarray
        Box-Cox power transformed array
    """

    try:
        box_cox_data, _ = scipy_stats.boxcox(arr, alpha=alpha, lmbda=lmbda)
    except ValueError as e:
        if const_policy == "propagate":
            return np.array([np.nan] * arr.shape[0])
        raise e

    return box_cox_data


def avg_moving_range(arr, nan_policy="omit"):
    """Calculate the average moving range (over 2-consecutive point1)

    Parameters
    ----------
    arr : array-like (1-D)
        Input array. Must be positive 1-dimensional.
    nan_policy : str, optional
        Value must be one of the following: {‘propagate’, ‘raise’, ‘omit’}
        Defines how to handle when input contains nan. The following options
        are available (default is ‘omit’):
        ‘propagate’: returns nan
        ‘raise’: throws an error
        ‘omit’: performs the calculations ignoring nan values

    Returns
    ----------
    np.ndarray, np.nan
        Average moving range. Returns NaN if arr is empty
    """

    arr = process_nan_policy(arr, nan_policy)
    if len(arr) == 0:
        return np.nan
    return np.mean(np.absolute(np.diff(arr)))


###################################################
# Stats related utilities
###################################################


def remove_nan(arr):
    """Remove indices from 1-D array with values of np.nan

    Parameters
    ----------
    arr : np.ndarray (1-D)
        Input array. Must be positive 1-dimensional.

    Returns
    ----------
    np.ndarray
        arr with NaN values deleted

    """
    return arr[~np.isnan(arr)]


def remove_const_column(arr):
    """Remove all columns with zero variance

    Parameters
    ----------
    arr : np.ndarray
        Input array (2-D)

    Returns
    ----------
    np.ndarray
        Input array with columns of a constant value removed
    """
    return arr[:, ~np.all(np.isnan(arr), axis=0)]


def is_nan_arr(arr):
    """Check if array has only NaN elements

    Parameters
    ----------
    arr : np.ndarray
        Input array

    Returns
    ----------
    bool
        True if all elements are np.nan
    """
    return np.all(np.isnan(arr))


def is_arr_constant(arr):
    """Determine if data by var_name is constant

    Parameters
    ----------
    arr : array-like
        Input array or object that can be converted to an array

    Returns
    ----------
    bool
        True if all values the same (i.e., no variation)
    """
    return np.all(arr == arr[0])


def process_nan_policy(arr, nan_policy):
    """Calculate the average moving range (over 2-consecutive point1)

    Parameters
    ----------
    arr : array-like (1-D)
        Input array. Must be positive 1-dimensional.
    nan_policy : str
        Value must be one of the following: {‘propagate’, ‘raise’, ‘omit’}
        Defines how to handle when input contains nan. The following options
        are available (default is ‘omit’):
        ‘propagate’: returns nan
        ‘raise’: throws an error
        ‘omit’: performs the calculations ignoring nan values

    Returns
    ----------
    np.ndarray, np.nan
        Input array evaluated per nan_policy
    """

    arr_no_nan = remove_nan(arr)
    if len(arr_no_nan) != len(arr):
        if nan_policy == "raise":
            msg = "NaN values are not supported for avg_moving_range"
            raise NotImplementedError(msg)
        if nan_policy == "propagate":
            return np.nan
        if nan_policy == "omit":
            return arr_no_nan
    return arr


def get_ordinal(n):
    """Convert number to its ordinal (e.g., 1 to 1st)

    Parameters
    ----------
    n : int
        Number to be converted to ordinal

    Returns
    ----------
    str
        the ordinal of n
    """
    return "%d%s" % (
        n,
        "tsnrhtdd"[(n // 10 % 10 != 1) * (n % 10 < 4) * n % 10 :: 4],
    )



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# test_utilities.py
"""unittest cases for utilities."""
#
# Copyright (c) 2020 Dan Cutright
# This file is part of DVHA-Stats, released under a MIT license.
#    See the file LICENSE included with this distribution, also
#    available at https://github.com/cutright/DVHA-Stats


import unittest
from os.path import join
import numpy as np
from numpy.testing import assert_array_equal, assert_array_almost_equal
from dvhastats import utilities

basedata_dir = join("tests", "testdata")
example_data = join(basedata_dir, "multivariate_data_small.csv")
example_data_nh = join(basedata_dir, "multivariate_data_small_no-header.csv")


class TestUtilities(unittest.TestCase):
    """Unit tests for Utilities."""

    def setUp(self):
        """Setup files and base data for utility testing."""
        self.data_path = example_data
        self.data_path_nh = example_data_nh

        data = [
            [np.nan, 48.1, 48.3, 65.1, 47.1, 49.9, 49.5, 48.9, 35.5, 44.5],
            [51.9, 44.3, 44.5, 58.7, 41.1, 43.9, 43.9, 44.5, 31.1, 40.5],
            [48.5, 38.5, 37.1, 53.9, 27.1, 40.7, 34.5, 41.7, 28.7, 37.7],
            [33.9, 21.5, 20.1, 48.7, 12.1, 38.5, 13.9, 33.7, 25.3, 28.1],
            [19.1, 10.9, 9.9, 42.3, 4.3, 36.5, 6.1, 16.9, 13.9, 13.3],
            [12.7, 4.7, 3.9, 31.3, 3.1, 29.1, 3.7, 4.9, 6.9, 4.7],
        ]
        keys = ["V%s" % (i + 1) for i in range(6)]

        self.expected_dict = {key: data[i] for i, key in enumerate(keys)}
        self.expected_dict_nh = {i: row for i, row in enumerate(data)}
        self.expected_arr = np.array(data).T
        self.expected_var_names = keys

    def test_csv_to_dict(self):
        """Test csv_to_dict"""
        data = utilities.csv_to_dict(self.data_path, dtype=float)
        self.assertEqual(data, self.expected_dict)

    def test_dict_to_array(self):
        """Test dict_to_array"""
        dict_data = utilities.csv_to_dict(self.data_path, dtype=float)
        arr = utilities.dict_to_array(dict_data)
        assert_array_equal(arr["data"], self.expected_arr)
        self.assertEqual(arr["var_names"], self.expected_var_names)

    def test_no_header_row(self):
        """Test import with no header row"""
        data = utilities.csv_to_dict(
            self.data_path_nh, dtype=float, header_row=False
        )
        self.assertEqual(data, self.expected_dict_nh)

    def test_apply_dtype(self):
        """Test the apply_dtype function"""
        test = utilities.apply_dtype("2.5", float)
        self.assertEqual(test, 2.5)
        test = utilities.apply_dtype(2.5, str)
        self.assertEqual(test, "2.5")
        test = utilities.apply_dtype(2.5, None)
        self.assertEqual(test, 2.5)


if __name__ == "__main__":
    import sys

    sys.exit(unittest.main())
